from itertools import chain
import re

from log import log
from shlex import quote, shlex  # shlex je teď můj


ORDER = 'kegncpamdxytzw~'
ALL_META = ('line', 'optional', 'structural', 'mwe', 'closing', 'empty',
            'begin', 'length', 'head', 'dep', 'edge')
META_ATTRS = ('head', 'dep')
FIXED_ATTRS = ('word', 'lemma', 'tag')


class Symbol(dict):
    def __init__(self, symbol=None, line=None, optional=None, structural=None,
                 mwe=None, closing=None, empty=None, begin=None, length=None,
                 head=None, dep=None, edge=None, **kwargs):
        if symbol is not None:
            if isinstance(symbol, str):
                symbols = read_rule(symbol)
                try:
                    symbol = next(symbols)
                except StopIteration:
                    raise ValueError('No symbol passed: "' + symbol + '"')
                # '•' neprojde
                extra = list(symbols)
                if extra:
                    # TODO: taky ValueError?
                    log.warning('Extra symbols passed, ignoring them: %s',
                                ' '.join(str(t) for t in extra))
            super().__init__(symbol)
            self.line = symbol.line
            self.optional = symbol.optional
            self.structural = symbol.structural
            self.mwe = symbol.mwe
            self.closing = symbol.closing
            self.empty = symbol.empty
            self.begin = symbol.begin
            self.length = symbol.length
            self.head = symbol.head
            self.dep = symbol.dep
            self.edge = symbol.edge
        else:
            super().__init__(**kwargs)
            self.line = line
            self.optional = optional
            self.structural = structural
            self.mwe = mwe
            self.closing = closing
            self.empty = empty
            self.begin = begin
            self.length = length
            self.head = head
            self.dep = dep
            self.edge = edge
        # TODO: ještě nevím, jestli nepovinné/prázdné fráze vždycky skrývat
        # self.hidden = False

    @property
    def end(self):
        return self.begin + self.length

    def __str__(self, show_tag=False):
        if self.structural:
            return self.stringify_structural()
        attrs = dict(self)
        phrase_head = attrs.pop('phrase', None)
        phrase_head = [(phrase_head, None)] if phrase_head else []
        optional = '?' if self.optional else ''

        if not show_tag:
            attrs.pop('tag', None)
        meta_attrs = [(attr, self.__dict__[attr]) for attr in META_ATTRS
                      if self.__dict__[attr]]
        fixed_attrs = [(attr, attrs[attr]) for attr in FIXED_ATTRS
                       if attr in attrs]
        short_attrs = [(attr, attrs[attr]) for attr in ORDER if attr in attrs]
        other_attrs = sorted((attr, val) for (attr, val) in attrs.items()
                             if attr not in FIXED_ATTRS and (
                                len(attr) > 1 or attr not in ORDER))
        span = ([('%s—%s' % (self.begin, self.end), None)]
                if self.begin is not None else [])
        # TODO: fakt je asi lepší shodu ukládat jako True
        attrs = (attr + '=' + escape(val) if val is not None else attr
                 for attr, val in chain(phrase_head, meta_attrs, fixed_attrs,
                                        short_attrs, other_attrs, span))
        return '({symbol})'.format(symbol=' '.join(attrs)) + optional

    def __repr__(self):
        return 'Symbol(' + escape(self) + ')'

    def update(self, symbol=None, **kwargs):
        if symbol is not None:
            super().update(symbol)
        if isinstance(symbol, Symbol):
            for meta in ALL_META:
                self.__dict__[meta] = symbol.__dict__[meta]
        for attr in dict(kwargs):
            if attr in ALL_META:
                self.__dict__[attr] = kwargs
                kwargs.pop(attr, None)

    # TODO: rozdělit/přesunout a zpřehlednit i v použití ve fundamentálním pr.!
    # přesunout třeba do Edge?
    def match(self, chart_edge):
        if chart_edge.current_symbol is not None:
            for attr, value in chart_edge.current_symbol.items():
                # log.debug('             trying to match %s=%s in %s', attr,
                #           value, self)
                try:
                    if value is None:
                        pass
                    elif attr not in self:
                        log.warning('not present  %s in %s', attr, self)
                    elif self[attr] is None:
                        log.warning('null value   %s=%s', attr, self[attr])
                    elif not re.fullmatch(value, self[attr]):
                        # log.warning('mismatch     %s=%s should be %s in %s '
                        #             'to match %s', attr, self[attr], value,
                        #             self, chart_edge)
                        return False
                except TypeError as te:
                    log.exception(te)
                    log.debug('%s %s %s', value, attr, self.get(attr))
                    raise
        else:
            log.warning('no symbol    cannot match %s to %s', self, chart_edge)
        for attr, indexes in sorted(chart_edge.rule.attr_matches.items()):
            value = None
            for index in indexes:
                if index == -1:
                    continue
                if index < chart_edge.position:
                    if attr not in chart_edge[index]:
                        log.warning('             %s not in %s', attr,
                                    chart_edge[index])
                        continue
                    value = chart_edge[index][attr]
                elif value is not None and index == chart_edge.position:
                    if attr not in self:
                        log.warning('not present  %s in %s', attr, self)
                    elif value != self[attr]:
                        log.error('mismatch     %s %s != %s in %s and %s',
                                  attr, self[attr], value, self, chart_edge)
                        return False
        return True

    def stringify_structural(self):
        closing = '/' if self.closing else ''
        tag = self.structural
        empty = '/' if self.empty else ''
        attrs = ''.join(
            ' {}="{}"'.format(attr, val) for attr, val in sorted(self.items()))
        return '<{closing}{tag}{attrs}{empty}>'.format(
            closing=closing, tag=tag, attrs=attrs, empty=empty)

    def vertical(self):
        return '{}\t{}\t{}'.format(self['word'], self['lemma'], self['tag'])


class Symbols(list):
    """Abstrakce pro pravou stranu pravidla/hrany (anebo její část)"""
    def __init__(self, iterable=None, position=None):
        self.position = position
        if isinstance(iterable, str):
            self.extend(read_rule(iterable))
        elif iterable is not None:
            super().__init__((Symbol(s) for s in iterable))
        else:
            super().__init__()
        if self and position is not None:
            self.parse_position()

    def __add__(self, value):
        if isinstance(value, Symbol):
            self.append(value)
        elif isinstance(value, str):
            self.extend(Symbols(value))
        else:
            return super().__add__(value)

    def __str__(self):
        if not self:
            return 'ε' if self.position is None else 'ε •'
        symbols = [str(symbol) for symbol in self]
        if self.position is not None:
            symbols.insert(self.position, '•')
        return ' '.join(symbols)

    def __repr__(self):
        return 'Symbols(' + escape(self) + ')'

    def parse_position(self):
        for index, symbol in enumerate(list(self)):
            if symbol == '•':
                if self.position is not None:
                    raise ValueError('Duplicated • in: ' + str(self))
                self.pop(index)
                self.position = index


def read_rule(rule):
    """
    Return a list of symbols, i.e. (non-)terminals from a string
    (which has already been split to the left/right side)
    """
    if rule.strip() == 'ε':
        return

    symbol = None  # not None ⇔ inside the symbol
    head = False  # expecting the phrase head (or nothing)
    attr = ''  # a = plus an attribute value may follow
    equal_sign = False  # we are expecting a value for an attribute after =
    closed = False  # the symbol has already ended with )

    for token in shlex(rule, posix=True):
        if token == '(' or (token == '•' and (symbol is None or closed)):
            if symbol is not None:
                yield symbol
                if not closed:  # TODO: allow ambiguous (a=( b=1) ?
                    raise ValueError('Unexpected "("')

            if token == '•':
                yield '•'
                symbol = None
                continue

            symbol = Symbol()
            head = closed = False
        elif token == '?':
            symbol.optional = True
        elif token == ')':
            if equal_sign:
                raise ValueError('Unterminated attribute ' + attr)
            elif attr:
                symbol[attr] = None  # není lepší pro shodové atributy True?

            attr = ''
            equal_sign = False
            closed = True
        elif closed:
            raise ValueError('Not expected outside a symbol: ' + token)
        elif not attr:
            if token == '=':
                raise ValueError('Unexpected "="')
            elif not head and token[0].isupper():
                head = symbol['phrase'] = token
                continue
            attr = token
        elif token == '=':
            equal_sign = True
        elif equal_sign:
            symbol[attr] = token
            attr = ''
            equal_sign = False
        else:
            symbol[attr] = None
            attr = token

    if symbol is not None:
        yield symbol
        if not closed:
            raise ValueError('Rule not terminated by ")"')


need_escaping = re.compile(r'[()=? \t\n]').search  # podle shlex._find_unsafe


def escape(s):  # začátek podle shlex.quote
    if not isinstance(s, str):
        s = str(s)
    if not s:
        return "''"
    elif need_escaping(s) is None:
        return s
    elif "'" not in s:
        return "'" + s + "'"
    elif '"' not in s:
        return '"' + s + '"'
    else:
        raise ValueError("Can't escape, the string includes both quotes: " + s)
